#!/usr/bin/python -OO
"""
--stefan huber, 2007-12
"""


import rect
import circle
from math import pi;
from math import atan2;
from math import cos;
from math import sin;
from math import sqrt;
import math

class arc:

		def __init__(self, center, radius, startangle, length):
				"""A arc with specific center and radius. Startangle is thought to be
				in mathemetical sense: 0 is in east-direction, growing CCW. The length
				is the length of the arc per radius, hence the delta of the angle and
				in the interval [-2pi,2pi] where positive numbers denote CCW arcs."""

				assert( radius >= 0 )
				assert( 0 <= startangle <= 2*pi )
				assert( -2*pi <= length <= 2*pi )

				#If oriented in CW -- make CCW
				if length < 0:
					length,startangle = -length,startangle+length

					if startangle < 0:
						startangle += 2*pi

				assert( startangle>=0 )
				assert( length>= 0)

				self.x, self.y = center
				self.radius = radius
				self.startangle = startangle
				self.length = length



		def __cmp__(self, obj):

				if not isinstance(obj, arc):
					return -1

				if (self.x, self.y, self.radius, self.startangle, self.length) < \
						(obj.x, obj.y, obj.radius, obj.startangle, obj.length):
					return -1;

				elif (self.x, self.y, self.radius, self.startangle, self.length) > \
						(obj.x, obj.y, obj.radius, obj.startangle, obj.length):
					return 1;

				else:
					return 0;

		def __hash__(self):
				return hash( hash(self.radius) + hash(self.x) + hash(self.y) + \
							hash(self.startangle) + hash(self.length) )

		def __repr__(self):
				return "arc[" + repr(self.x) + ", " + repr(self.y) + \
								", "+ repr(self.radius) + ", phi: "+ repr(self.startangle) + \
								" + "+ repr(self.length) + "]"

		def circle(self):
				return circle.circle( (self.x, self.y), self.radius )

		def clone(self):
				return arc((self.x, self.y), self.radius, self.startangle, self.length)

		def start(self):
				phi = self.startangle

				dx = self.radius*cos(phi)
				if sin(phi)>=0:
					dy = sqrt(self.radius**2 - dx**2)
				else:
					dy = -sqrt(self.radius**2 - dx**2)

				dy = self.radius*sin(phi)
				
				return ( self.x + dx, self.y + dy )

		def end(self):
				phi = self.startangle + self.length

				dx = self.radius*cos(phi)
				if sin(phi)>=0:
					dy = sqrt(self.radius**2 - dx**2)
				else:
					dy = -sqrt(self.radius**2 - dx**2)
				
				dy = self.radius*sin(phi)
				
				return ( self.x + dx, self.y + dy )


		def endangle(self):
				if self.startangle + self.length >= 2*pi:
					return self.startangle + self.length - 2*pi
				else:
					return self.startangle + self.length


		def diffcenter(self, v):
				vx, vy = v
				return vx-self.x, vy-self.y

		def startv(self):
				return self.diffcenter(self.start())

		def endv(self):
				return self.diffcenter(self.end())

		def normstartv(self):
				"""startv 90 deg CCW"""
				vx, vy = self.startv()
				return -vy, vx

		def normendv(self):
				"""endv 90 deg CW"""
				vx, vy = self.endv()
				return vy, -vx


		def boundingBox(self):
				"""Get the bounding box"""

				xs,ys = self.start()
				xe,ye = self.end()
				radius = self.radius

				if self.containsAngle(pi):
					left = self.x-radius
				else:
					left = min(xs,xe)

				if self.containsAngle(0):
					right = self.x+radius
				else:
					right = max(xs,xe)

				if self.containsAngle(pi/2):
					top = self.y+radius
				else:
					top = max(ys,ye)

				if self.containsAngle(3*pi/2):
					bot = self.y-radius
				else:
					bot = min(ys,ye)

				return rect.rect(left,right,bot,top)


		def containsAngle( self, angle ):
				"""Test if ray from center in direction of angle hits arc"""

				assert( 0 <= angle <= 2*pi )

				diff = angle-self.startangle
				if diff < 0:
					diff += 2*pi

				if 0 <= diff <= self.length:
					return True
				else:
					return False


		def ciContainsPnt( self, p, eps=0.0 ):
				"""Test if point is contained in cone of influence"""


				#if not self.circle().containsPnt(p,eps):
				#	return False

				vx,vy = self.diffcenter(p)
				svx, svy = self.normstartv()
				evx, evy = self.normendv()
				
				if self.length <= pi:
					if svx*vx + svy*vy >= -eps and \
						evx*vx + evy*vy >= -eps: 
						return True
					else:
						return False

				else:
					if svx*vx + svy*vy >= -eps or \
						evx*vx + evy*vy >= -eps: 
						return True
					else:
						return False


		def containedInBox(self, rect, eps=0.0):
				"""Test if closure of arc is contained in box"""				
				return self.boundingBox().containedInBox(rect,eps)


		def intersectsBox(self, rect, eps=0.0):
				"""Test if arc intersects with filled box."""

				#The bounding circle
				c = self.circle()
				
				#Does not intersect the bounding box!
				if not self.boundingBox().intersectsBox(rect,eps):
					return False

				#Does not intersect the circle!
				#if not c.intersectsBox(rect, -eps, False):
				#	return False

				#The whole circle is contained in box
				if self.boundingBox().containedInBox(rect, eps):
					return True

				
				#Compute the intersections with box and test if they belong to arc
				for p in c.compIntersectRect(rect, eps):
					if self.ciContainsPnt(p, eps):
						return True

				return False


		def compIntersect(self, obj, eps=0.0):
				"""Compute intersection of self with obj"""

				c = self.circle()

				if isinstance(obj, rect.rect):
					return filter( lambda p: self.ciContainsPnt(p, -eps), c.compIntersectRect(obj,eps))

				elif isinstance(obj, circle.circle):
					return filter( lambda p: self.ciContainsPnt(p, -eps), c.compIntersectCircle(obj,eps))

				elif isinstance(obj, arc):
					return filter( lambda p: self.ciContainsPnt(p, -eps) and obj.ciContainsPnt(p, -eps), \
									c.compIntersectCircle(obj.circle(),eps))
				else:
					raise Error("obj has unkown type")


		def getSplittedBy(self, other, eps=0.0):
				"""Get arcs when splitting self by other"""

				def normalize(x):
					while x>=1.99*pi:
						x -= 2*pi
					while x<0:
						x += 2*pi
					return x
				
				x,y,rad = self.x, self.y, self.radius

				#Compute intersections
				ins = self.compIntersect(other, eps)

				#No intersection
				if len(ins) == 0:
					return [self.clone()]


				#Compute arc length from startangle
				angles = map(normalize, [ atan2(py-y, px-x)-self.startangle for px,py in ins ])
				angles.sort()
				#angles =  filter(lambda x: x<self.length, angles) + [ self.length ]
				angles +=  [ self.length ]
		
				#Construct the arcs...
				arcs = []
				oldangle = 0
				for angle in angles:
				
					length = angle-oldangle
					start = normalize(oldangle + self.startangle)
					assert( 0 < length < self.length )

					if length > eps:
						assert( math.fabs(normalize(start+length) - normalize(angle+self.startangle) )<1e-11 )

						arcs += [arc( (x,y), rad, start, length)]
						oldangle = angle

				return arcs



