#!/usr/bin/python -OO
"""
--stefan huber, 2007-12
"""


import rect
import arc
from math import sqrt
from math import pi
from math import atan2
from math import fabs
import math


class circle:


		def __init__(self, center, radius):
				self.x, self.y = center
				self.radius = radius

		def __cmp__(self, obj):

				if not isinstance(obj, circle):
					return -1

				if (self.x, self.y, self.radius) < (obj.x, obj.y, obj.radius):
					return -1;

				elif (self.x, self.y, self.radius) > (obj.x, obj.y, obj.radius):
					return 1;

				else:
					return 0;

		def __hash__(self):
				return hash( hash(self.radius) + hash(self.x) + hash(self.y))

		def __repr__(self):
				return "circle[" + repr(self.x) + ", " + repr(self.y) + \
								", "+ repr(self.radius) + "]"

		def clone(self):
				return circle((self.x, self.y), self.radius)


		def north(self):
				return self.x, self.y+self.radius

		def south(self):
				return self.x, self.y-self.radius

		def west(self):
				return self.x-self.radius, self.y

		def east(self):
				return self.x+self.radius, self.y
	

		def boundingBox(self):
				"""Get the bounding box"""
				x,y,radius = self.x, self.y, self.radius
				return rect.rect( x-radius, x+radius, y-radius, y+radius)


		def containsPnt( self, p, eps=0.0 ):
				"""Test if closure of circle contains point"""

				px, py = p
				cx, cy = self.x, self.y

				if (px-cx)**2 + (py-cy)**2 <= self.radius**2 + 2*self.radius*eps + eps**2 :
					return True

				else:
					return False


		def containedInBox(self, rect, eps=0.0):
				"""Test if closure of circle is contained in box"""
				
				if rect.containsPnt(self.north(), eps) and \
					rect.containsPnt(self.south(), eps) and \
					rect.containsPnt(self.west(), eps) and \
					rect.containsPnt(self.east(), eps):
					return True

				else:
					return False


		def intersectsBox(self, rect, eps=0.0, filled=False):
				"""Test if circle intersects with filled box. if filled is set
				the circle is thought to be a disk"""

				
				#Circle intersects box iff
				#  - bounding box intersects rect and
				#  	- a corner point is in circle or
				#   - left and right of rect are seperated by north-south line or
				#   - top and bot of rect are seperated by west-east line

				if not self.boundingBox().intersectsBox(rect,eps):
					return False

				#If circle is not filled and no corner point of rect
				#is in circle the rect does not intersect the circle-line
				if not filled and \
					self.containsPnt(rect.lt(), -eps) and \
					self.containsPnt(rect.rt(), -eps) and \
					self.containsPnt(rect.lb(), -eps) and \
					self.containsPnt(rect.rb(), -eps) :
					return False

				#So here the bounding box intersects with rect

				#Circle contains a corner point of rect
				if self.containsPnt(rect.lt(), eps) or \
					self.containsPnt(rect.rt(), eps) or \
					self.containsPnt(rect.lb(), eps) or \
					self.containsPnt(rect.rb(), eps) :
					return True

				#rect is hit by north-south line or by east-west line
				if rect.left-eps < self.x < rect.right+eps or \
					rect.bot-eps < self.y < rect.top+eps:
					return True

				else:
					return False



		def compIntersectRect(self, rect, eps=0.0):
				"""Retrieves a list of intersection points"""

				#No intersection
				if not self.intersectsBox(rect):
					return []

				res = []
				x,y = self.x, self.y

				rad = self.radius
				left = rect.left-x
				right = rect.right-x
				top = rect.top-y
				bot = rect.bot-y


				if left**2 <= (rad+eps)**2:

						if left**2 >= (rad-eps)**2:
							res += [ (x-rad,y) ]
						else:
							res += [ (x+left, y + sqrt(rad**2-left**2)), \
									 (x+left, y - sqrt(rad**2-left**2))]

				if right**2 <= (rad+eps)**2:

						if right**2 >= (rad-eps)**2:
							res += [ (x+rad,y) ]
						else:
							res += [ (x+right, y + sqrt(rad**2-right**2)), \
									 (x+right, y - sqrt(rad**2-right**2))]

				if bot**2 <= (rad+eps)**2:

						if bot**2 >= (rad-eps)**2:
							res += [ (x,y-rad) ]
						else:
							res += [ (x+sqrt(rad**2-bot**2), y+bot), \
									 (x-sqrt(rad**2-bot**2), y+bot)]

				if top**2 <= (rad+eps)**2:

						if top**2 >= (rad-eps)**2:
							res += [ (x,y+rad) ]
						else:
							res += [ (x+ sqrt(rad**2-top**2), y+top), \
									 (x- sqrt(rad**2-top**2), y+top)]

				#Filter those which do not belong to box
				res = filter( lambda p: rect.containsPnt(p,eps), res)

				return res



		def compIntersectCircle(self, circ, eps=0.0):
				"""Computing the intersection between two circles"""

				x1, y1, r1 = self.x, self.y, self.radius
				x2, y2, r2 = circ.x, circ.y, circ.radius

				#Squared distance between centers
				cc = (x1-x2)**2 + (y1-y2)**2

				#Concentric circles
				if cc <= eps:
					return []

				#Distance of projection of intersection on centerline to c1
				a = (cc + r1**2 - r2**2)/(2*sqrt(cc))
				#Squared ortho distance of intersection to centerline
				h = r1**2 - a**2

				if h < -eps:
					return []

				#Projection on center-line
				px, py = x1 + a*(x2-x1)/sqrt(cc), y1 + a*(y2-y1)/sqrt(cc)

				if h <= eps:
					return [ (px,py) ]

				else:
					ox, oy = (y1-y2)*sqrt(h)/sqrt(cc), (x2-x1)*sqrt(h)/sqrt(cc)
					return[ (px+ox, py+oy), (px-ox, py-oy)]

				

		def compIntersect( self, obj, eps=0.0):
				"""Compute the intersections of circle with obj"""

				if isinstance(obj, rect.rect):
					return self.compIntersectRect(obj, eps)

				elif isinstance(obj, circle):
					return self.compIntersectCircle(obj, eps)

				elif isinstance(obj, arc.arc):
					return filter( lambda p: obj.ciContainsPnt(p, -eps), self.compIntersectCircle(obj.circle(),eps))

				else:
					raise Error("obj has unknown type");


		def getSplittedBy(self, other, eps=0.0):
				"""Get arcs when splitting self by other"""

				def normalize(x):
					while x>=2*pi:
						x -= 2*pi
					while x<0:
						x += 2*pi
					return x

				
				x,y,rad = self.x, self.y, self.radius

				#Compute intersections
				ins = self.compIntersect(other, eps)

				#No intersection
				if len(ins) == 0:
					return [self.clone()]

				#Compute angles
				angles = [ atan2(py-y, px-x) for px,py in ins ]
				first = normalize(angles[0])
				angles = map( lambda x: normalize(x-first), angles)
				#Make full circle
				angles[0] = 2*pi
				angles.sort()

				#Only one split point -> 2*pi arc
				if len(angles) == 1:
					return [ arc.arc( (x,y), rad, first, 2*pi) ]


				#Construct the arcs...
				arcs = []
				oldangle = 0
				for angle in angles:
				
					length = angle-oldangle
					start = normalize(oldangle + first)

					if length > eps:
						assert( math.fabs(normalize(start+length) - normalize(angle+first) )<1e-11 )

						arcs += [arc.arc( (x,y), rad, start, length)]
						oldangle = angle


			
				return arcs




