#!/usr/bin/python 
"""
--stefan huber, 2007-12
"""

from PyQt4 import QtGui, QtCore

from geomhashtable import geomhashtable
from point import point
from rect import rect
from circle import circle
from arc import arc
from math import pi


class QGeomhashtableWidget (QtGui.QWidget):
		"""QT Widget carrying the geom hash table"""

		bgcolor = QtGui.QColor("black")
		objcolor = QtGui.QColor(0,255,0)
		pntcolor = QtGui.QColor(255,0,0)
		gridcolor = QtGui.QColor(0,0,128)
		gridfilledcolor = QtGui.QColor(64,64,255,64)

		zoomfactor = None
		pixmap = None


		def __init__(self, ght, zoomfactor, *args):

				self.ght = ght
				self.first = True

				apply( QtGui.QWidget.__init__, (self,) + args)
				self.zoomFactorChanged(zoomfactor)


		def paintEvent(self, ev):			

				p = QtGui.QPainter()
				p.begin(self)
				if self.pixmap != None:
					p.drawPixmap(0, 0, self.pixmap)
					pass
				p.end()


		def wheelEvent(self, ev):

				fact = self.zoomfactor* 2**(ev.delta()/120)
				self.zoomFactorChanged( max(16,min(1<<14, fact)) )

		
		def zoomFactorChanged(self, factor ):
				"""Zoom factor changed"""

				print "repainting for zoom-factor", factor,

				if factor == self.zoomfactor:
					return

			
				self.zoomfactor = factor
				self.pixmap = QtGui.QPixmap(int(self.zoomfactor), int(self.zoomfactor))

				p = QtGui.QPainter()
				p.begin(self.pixmap)

				self.pixmap.fill( self.bgcolor )					
				self.paintGrid(p)
					
				p.setRenderHint(QtGui.QPainter.Antialiasing)
				self.paintObjects(p)

				p.end()

				self.setMinimumSize( int(self.zoomfactor), int(self.zoomfactor))
				self.resize( int(self.zoomfactor), int(self.zoomfactor))
				
				print "."


		def paintObjects(self, p ):
				"""Paint all objects on QPainter p"""

				pen = QtGui.QPen( self.objcolor, 1)
				p.setPen(pen)
				p.setBrush(QtGui.QColor(0,0,0,0));

				size = self.zoomfactor

				for obj in self.ght:
					p.setPen(pen)

					if isinstance(obj, circle):
						x,y = obj.boundingBox().lt()
						r = obj.radius
						p.drawEllipse(x*size, (1-y)*size, 2*r*size, 2*r*size)

					elif isinstance(obj, rect):
						p.drawRect(QtCore.QRect( rect.left*size, (1-rect.bot)*size, \
										rect.width()*size, rect.height()*size ))

					elif isinstance(obj, arc):
						x,y = obj.circle().boundingBox().lt()
						r = obj.radius
						p.drawArc(x*size, (1-y)*size, 2*r*size, 2*r*size, 16*180/pi*obj.startangle, 16*180/pi*obj.length)

						pen2 = QtGui.QPen( self.pntcolor, 1)
						p.setPen(pen2)
						x,y = obj.start()
						x,y = x, y
						p.drawRect(x*size-2, (1-y)*size-2, 3, 3)
						x,y = obj.end()
						x,y = x, y
						p.drawRect(x*size-2, (1-y)*size-2, 3, 3)

					elif isinstance(obj, point):
						x,y = obj.coord()
						p.drawLine(x*size-2, (1-y)*size-2, x*size+1, (1-y)*size+1)
						p.drawLine(x*size-2, (1-y)*size+1, x*size+1, (1-y)*size-2)




		def paintGrid(self, p):
				"""Paint the grid"""

				p.setPen(self.gridcolor)
				size = self.zoomfactor

				for i in xrange(0, self.ght.resx+1):
					x = float(i)/self.ght.resx*size
					p.drawLine(x, 0, x, size)

				for j in xrange(0, self.ght.resy+1):
					y = float(j)/self.ght.resy*size
					p.drawLine(0, y, size, y)


				p.setPen(QtGui.QColor(0,0,0,0))
				p.setBrush(self.gridfilledcolor);

				for (i,j) in self.ght.getGridIterator():

						rect = self.ght.rectOfGridIdx(i,j)
						x,y = rect.lt()
						qrect = QtCore.QRect(x*size, (1-y)*size, rect.width()*size, rect.height()*size)

						for k in xrange(0, min(10,len(self.ght.table[i,j]))):
							p.drawRect(qrect)


			


class MainWindow(QtGui.QMainWindow):

		def __init__(self, ght, zoomfactor, *args):

			apply( QtGui.QMainWindow.__init__, (self,) + args)
			
			self.sa = QtGui.QScrollArea(self)
			self.painting = QGeomhashtableWidget(ght, zoomfactor, self.sa)

			self.sa.setWidgetResizable(True)
			self.sa.setWidget(self.painting)

			self.setCentralWidget(self.sa)
			self.resize(int(zoomfactor)+10, int(zoomfactor)+10)
						



